# author: Volfin, Aspadm

import bpy
import os
import sys
import mathutils
import math
import platform
import imp
import csv
import struct
from bpy_extras.io_utils import unpack_list
from struct import *
from mathutils import Vector, Matrix, Quaternion, Euler
import time
import random

objects_ptr = [];  # list of object data structures.
header_size = 0 #init

def to_blender_matrix(bl_mat):
    mat = mathutils.Matrix()
    for i, col in enumerate(bl_mat):
        for k, val in enumerate(col):
            mat[k][i] = val
    return mat

def import_mesh(context,randomize_colors,import_vertcolors,use_layers,mesh_scale,msh_count,block_offset):
    
    global model_name,header_size
    SWAP_YZ_MATRIX = mathutils.Matrix.Rotation(math.radians(0.0), 4, 'X')

    # lets just do this to be sure    
    if bpy.ops.object.mode_set.poll():
        bpy.ops.object.mode_set(mode='OBJECT')
    #else:
    #    print("mode_set() context is incorrect, current mode is", bpy.context.mode)

    # stage mesh
    filehandle.seek(block_offset,0) # SEEK_SET

    #Block 1
    print("Block1 Offset:"+str(filehandle.tell()))
    BH(1)
    FormatFlag=BH(1)[0]
    print("FormatFlag ="+str(FormatFlag))
    SkipFlag = BH(1)[0]
    print("SkipFlag ="+str(SkipFlag))
    UVFlag=BH(1)[0]
    print("UV Flag:"+str(UVFlag))

    model_name2=model_name+"-"+str(msh_count)
    me = bpy.data.meshes.new(model_name2)
    me_obj = bpy.data.objects.new(model_name2,me)
    me_obj.select = True # makes selected

    objects_ptr.extend([me_obj]) #store for later
    context.scene.objects.link(me_obj)
    context.scene.objects.active = me_obj # makes active

    # read in mesh data
    vertex=[]
    normals=[]
    uvs=[]
    UV2=[]
    UV3=[]
    UV4=[]
    colors=[]
    colors2=[]
    bone_list=[]
    w_count=[]
    m_weights=[]
    g_indices=[]
    faces=[]
    face_idx=[]
    alt_format = False

    #skip junk
    BI(11)
    next_offset = BI(1)[0] + header_size
    BI(2) #junk

    # mesh scale per axis
    x_scalefactor = Bf(1)[0]
    y_scalefactor = Bf(1)[0]
    z_scalefactor = Bf(1)[0]

    BI(5) #junk
    # UV adjustment data
    U_SCALE = Bf(1)[0]
    V_SCALE = Bf(1)[0]
    U_OFFSET = Bf(1)[0]
    V_OFFSET = Bf(1)[0]
  
    material_count = 0

    #kludgy way to detect alternate model format
    if (x_scalefactor == 0.0 or y_scalefactor == 0.0 or z_scalefactor == 0.0):
        alt_format=True
        x_scalefactor=1.0
        y_scalefactor=1.0
        z_scalefactor=1.0
    
    filehandle.seek(next_offset,0) # SEEK_SET

    #Block 2
    next_offset = BI(1)[0] + header_size
    filehandle.seek(next_offset,0) # SEEK_SET

    #Block 3
    print("Block3 Offset:"+str(filehandle.tell()))
    ID1=BB(1)[0]
    ID2=BB(1)[0]
    print("ID1:"+str(ID1)+" ID2:"+str(ID2))
    BH(1)
    next_offset3 = BI(1)[0] + header_size# not used
    if alt_format:
        BI(2) #unverfied
    else:
        BI(11)
    vert_count=BI(1)[0]
    vert_offset=BI(1)[0] + header_size
    face_count=BI(1)[0]
    #if not alt_format:
    #    BI(1)
    face_offset=BI(1)[0] + header_size
    mat_offset = BI(1)[0] + header_size
    wght_offset = BI(1)[0] + header_size

    if wght_offset >= 0xFFFF:
        wght_offset = next_offset

    print("Vert count:"+str(vert_count))
    print("Face count:"+str(face_count))

    print("mat_offset:"+str(mat_offset))
    print("wght_offset:"+str(wght_offset))
    print("next_offset:"+str(next_offset))

    #vertices
    filehandle.seek(vert_offset,0) # SEEK_SET    
    print("Vertice offset:"+str(filehandle.tell()))
    # vert_block
    print("Mesh #"+str(msh_count)+" Alt_format:"+str(alt_format)+" ID1:"+str(ID1))
    
    for x in range(0,vert_count):

        if (alt_format):
            if ID1 == 0:
                tmp=Bf(6)
                vertex.extend([(tmp[0],tmp[1],tmp[2])])
                uvs.extend([(tmp[4], tmp[5])])
                #normals.extend([tmp[5],tmp[6],tmp[7]])
                #BI(1) #unknown data
                #print("verts:"+str((tmp[0],tmp[1],tmp[2])))
            elif ID1 == 1:
                tmp=BF(8)
                vertex.extend([(tmp[0],tmp[1],tmp[2])])
                uvs.extend([(tmp[3]+0.5, tmp[4]+0.5)])
                normals.extend([tmp[5],tmp[6],tmp[7]])
                BI(2) #unknown data
                #print("verts:"+str((tmp[0],tmp[1],tmp[2])))
            elif ID1 == 2:
                tmp=BF(8)
                vertex.extend([(tmp[0],tmp[1],tmp[2])])
                uvs.extend([(tmp[3]+0.5, tmp[4]+0.5)])
                normals.extend([tmp[5],tmp[6],tmp[7]])
                BI(3) #unknown data
                #print("verts:"+str((tmp[0],tmp[1],tmp[2])))
            elif ID1 == 3:
                tmp=BF(8)
                vertex.extend([(tmp[0],tmp[1],tmp[2])])
                uvs.extend([(tmp[3]+0.5, tmp[4]+0.5)])
                normals.extend([tmp[5],tmp[6],tmp[7]])
                BI(6) #unknown data
                #print("verts:"+str((tmp[0],tmp[1],tmp[2])))
            elif ID1 == 4:
                tmp=BF(8)
                vertex.extend([(tmp[0],tmp[1],tmp[2])])
                uvs.extend([(tmp[3]+0.5, tmp[4]+0.5)])
                normals.extend([tmp[5],tmp[6],tmp[7]])
                BI(4) #unknown data
                #print("verts:"+str((tmp[0],tmp[1],tmp[2])))
            elif ID1 == 5:
                tmp=BF(8)
                vertex.extend([(tmp[0],tmp[1],tmp[2])])
                uvs.extend([(tmp[3]+0.5, tmp[4]+0.5)])
                normals.extend([tmp[5],tmp[6],tmp[7]])
                BI(5) #unknown data
                #print("verts:"+str((tmp[0],tmp[1],tmp[2])))
            elif ID1 == 6:
                tmp=BF(8)
                vertex.extend([(tmp[0],tmp[1],tmp[2])])
                uvs.extend([(tmp[3]+0.5, tmp[4]+0.5)])
                normals.extend([tmp[5],tmp[6],tmp[7]])
                BI(8) #unknown data
                #print("verts:"+str((tmp[0],tmp[1],tmp[2])))
            elif ID1 == 7:
                tmp=BF(8)
                vertex.extend([(tmp[0],tmp[1],tmp[2])])
                uvs.extend([(tmp[3]+0.5, tmp[4]+0.5)])
                normals.extend([tmp[5],tmp[6],tmp[7]])
                BI(5) #unknown data
                #print("verts:"+str((tmp[0],tmp[1],tmp[2])))
            #else:
            # let it fail so i know.
            #    tmp=Bf(5)
            #    vertex.extend([(tmp[0],tmp[1],tmp[2])])
            #    uvs.extend([(tmp[3]+0.5, tmp[4]+0.5)])
            #    #print("else2 verts:"+str((tmp[0],tmp[1],tmp[2])))  
            
        else:
            if SkipFlag == 4096 or SkipFlag == 4099:
                tmp=Bf(3)
                vertex.extend([(tmp[0],tmp[1],tmp[2])])
            elif SkipFlag == 4098:
                tmp=Bf(4)
                vertex.extend([(tmp[0],tmp[1],tmp[2])])               
            else:
                tmp=BF(4)
                vertex.extend([(tmp[0],tmp[1],tmp[2])])
                

    # fudge a flag for UV block size
    
    uv_block_size = (mat_offset - filehandle.tell()) // vert_count
    if uv_block_size <= 0:
        uv_block_size = (wght_offset - filehandle.tell()) // vert_count

    if SkipFlag != 4096 and SkipFlag != 4098 and SkipFlag != 4099 and SkipFlag != 0:
        print ("Adjusting UV Block Size for Skipped data")
        uv_block_size = uv_block_size - 8 # 8 byte entries * vert_count / vert_count

    print("UV Block Size is:"+str(uv_block_size))
    print("UVs offset:"+str(filehandle.tell())+" Mat offset:"+str(mat_offset))
    
    if not alt_format:

        # not sure this works consistently, but ...  
        if SkipFlag == 4099 or SkipFlag == 4096:
            print('using alt UV format 2')
            for x in range(0,vert_count):
                #if UVFlag == 0x03 or UVFlag == 0x0C or UVFlag == 0x0F or UVFlag == 0x10 or UVFlag == 0x1C or UVFlag == 0x1F or UVFlag == 0xFF:
                    if uv_block_size == 20:
                        tmp=Bh(10)
                        uvs.extend([(((tmp[8]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[9]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                    elif uv_block_size == 24:
                        tmp=Bh(12)
                        uvs.extend([(((tmp[8]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[9]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                        UV2.extend([(((tmp[10]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[11]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                    elif uv_block_size == 28:
                        tmp=Bh(14)
                        uvs.extend([(((tmp[8]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[9]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                    elif uv_block_size == 32:
                        tmp=Bh(16)
                        uvs.extend([(((tmp[8]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[9]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                    else:
                        print('Block size '+str(uv_block_size)+' not found.')
        elif SkipFlag == 4098:
            print('using alt UV format 3')
            for x in range(0,vert_count):
                    if uv_block_size == 20:
                        tmp=Bh(10)
                        uvs.extend([(((tmp[4]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[5]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
        elif SkipFlag == 0x2:
            print ('using Aspadm UV format')
            if uv_block_size == 16:
                for x in range(0,vert_count):
                    tmp=Bh(12)
                    uvs.extend([(((tmp[6]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[7]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                    UV2.extend([(((tmp[10]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[11]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
        elif SkipFlag != 0x0:
            BI(vert_count*2) #8 bytes per
            print('using alt UV format')
            for x in range(0,vert_count):
                if uv_block_size == 20:
                    tmp=Bh(10)
                    uvs.extend([(((tmp[6]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[7]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                elif uv_block_size == 24: 
                    tmp=Bh(12)
                    uvs.extend([(((tmp[6]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[7]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                    UV2.extend([(((tmp[10]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[11]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                elif uv_block_size == 28:
                    tmp=Bh(14)
                    uvs.extend([(((tmp[6]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[7]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                else:
                    print('Block size '+str(uv_block_size)+' not found.')
        else:
            print("using default UV format")
            for x in range(0,vert_count):                
                if uv_block_size == 20:
                    tmp=Bh(10)
                    uvs.extend([(((tmp[8]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[9]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                elif uv_block_size == 24:
                    tmp=Bh(12)
                    uvs.extend([(((tmp[8]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[9]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                    UV2.extend([(((tmp[10]/32768.0)*U_SCALE)+U_OFFSET, 1.0-(((tmp[11]/32768.0)*V_SCALE)+V_OFFSET))]) # v axis was flipped
                else:
                    print('Block size '+str(uv_block_size)+' not found.')

    ####################################################################
         
    print("vertex array length:"+str(len(vertex)))
    print("normal array length:"+str(len(normals)))
    print("weights array length:"+str(len(m_weights)))
    print("index array length:"+str(len(g_indices)))
    print("UV1 array length:"+str(len(uvs)))
    print("UV2 array length:"+str(len(UV2)))
    print("UV3 array length:"+str(len(UV3)))
    print("Colors array length:"+str(len(colors)))

    #########################################
    ## FACES
    #########################################

    filehandle.seek(face_offset,0) # SEEK_SET
    print("Faces offset:"+str(filehandle.tell()))
    print("count / 3:"+str(face_count//3))
    
    # face block  
    for x in range(0,face_count//3):
        tmp=BH(3)
        faces.extend([(tmp[0],tmp[1],tmp[2])])
        face_idx.extend([0]) #initialize for fill-in later

    print("faces array length:"+str(len(faces)))


    try:
        me.from_pydata(vertex,[],faces)
        me.vertices.foreach_set("normal",normals) # add normals
    except:
        pass

    # set to smooth shaded	
    bpy.ops.object.shade_smooth()

    if use_layers:
        # place mesh on incremental layer, wrap back around if more than 20 layers
        
        me_obj.layers[msh_count%20] = True

        #wipe other layers
        for i in range(20):
            # if there's ever more than 20 meshes, we are in trouble. But i doubt that happens.
            me_obj.layers[i] = (i == msh_count)
    
    ###############################################
    ## Material Data
    ###############################################
    #material_count = 0 #BI(1)[0]

    # for now, skip past materials, will deal with later.
    if material_count > 0:
        filehandle.seek(material_offset+(material_count*4),0) # SEEK_SET ; skip past index numbers.
        start_index = 0; #start indexing from 0
        end_offset = 0;
        for x in range(0,material_count):
            matName = "Material_"+str(msh_count)+"_"+str(x)
            start_index = end_offset #old end is new start
            
            end_offset=BI(1)[0]//3 # range data for material

            mat = bpy.data.materials.new(matName)
            bpy.ops.object.material_slot_add()
            me_obj.material_slots[x].material = mat
            print("setting materials slot ",x," to ",mat)
            print("START:"+str(start_index)+" END:"+str(end_offset))

            for j in range(start_index,end_offset): # Start number, count
                #print("assigning material ",x," to index ",j)
                face_idx[j]=x

            if randomize_colors:
                mat.diffuse_color = randomColor()
            #mat.specular_color = material.specular()
            #mat.mirror_color = material.emissive()
            #mat.alpha = material.transparency_factor()
            #mat.specular_intensity = material.shininess()
        
            # texture
            #for k, texture in enumerate(material.texture_list()):    
            #    try:
            #        tex = mat.texture_slots.create(k)
            #        tex.texture = bpy.data.textures.new(texture.name(), type='IMAGE')
            #        tex.texture_coords = 'UV'
            #        tex.use = True
            #        tex.use_map_color_diffuse = True
            #        fbx_base_path, fbx_file_name = os.path.split(file)
            #        fbx_texture_file_path = texture.file_name()
            #        fbx_garbage_path, fbx_texture_name = os.path.split(fbx_texture_file_path)
            #        base_norm = os.path.normpath(fbx_base_path)
            #        garbage_norm = os.path.normpath(fbx_garbage_path)
            #        if base_norm in garbage_norm and os.path.exists(fbx_texture_file_path):
            #            texture_file_path = fbx_texture_file_path
            #        else:
            #            texture_file_path = bpy.path.abspath(os.path.join(fbx_base_path, fbx_texture_name))
            #        texture_file_path = os.path.normpath(texture_file_path)
            #        if os.path.exists(texture_file_path):
            #            tex.texture.image = bpy.data.images.load(texture_file_path)
            #    except:
            #        pass

            # read in some (as of now) unknown values

 
    # material index
    #print("Face_IDX:"+str(face_idx))
    me.polygons.foreach_set("material_index",face_idx) # should set multiple mat indexes.
    
    # post material data 
    bone_list_len = 0 #BI(1)[0]

    # may have bone list
    if bone_list_len != 0:
        for x in range(0,bone_list_len):
            bone_header=BI(7)
            name = file_str(bone_header[6]-1)
            #print("bone Name:",name," index:",x)
            bone_list.extend([name]) # bone_header contains name len + ending null
            BB(1) # read off the null terminator
            BI(25) # read off other data for now (bone location and rotation data)
        

    # and done with this model...
    

    ###############################################
    if face_count > 0: # skip all this if no mesh

    ###############################################
    # Do UVs
    ###############################################
        if len(uvs) != 0:
            me.uv_textures.new(name='UV_0')
            uv_data = me.uv_layers[0].data
            for i in range(len(uv_data)):
                uv_data[i].uv = uvs[me.loops[i].vertex_index]
        if len(UV2) != 0:
            me.uv_textures.new(name='UV_1')
            uv_data = me.uv_layers[1].data
            for i in range(len(uv_data)):
                uv_data[i].uv = UV2[me.loops[i].vertex_index]
        if len(UV3) != 0:
            me.uv_textures.new(name='UV_2')
            uv_data = me.uv_layers[2].data
            for i in range(len(uv_data)):
                uv_data[i].uv = UV3[me.loops[i].vertex_index]
        if len(UV4) != 0:
            me.uv_textures.new(name='UV_3')
            uv_data = me.uv_layers[3].data
            for i in range(len(uv_data)):
                uv_data[i].uv = UV4[me.loops[i].vertex_index]

    ###############################################
    # Do Vertex Color
    ###############################################

        if len(colors) != 0 and import_vertcolors:
            me.vertex_colors.new(name='Color_Data')
            color_data = me.vertex_colors[0].data
            for i in range(len(color_data)):
                color_data[i].color = colors[me.loops[i].vertex_index]
        if len(colors2) != 0 and import_vertcolors:
            me.vertex_colors.new(name='Color_Data2')
            color_data = me.vertex_colors2[0].data
            for i in range(len(color_data)):
                color_data[i].color = colors2[me.loops[i].vertex_index]


    ###############################################
    # Finalize
    ###############################################

        # finalize mesh
        me.update(calc_edges=True)

        #scale
        scale_matrix = mathutils.Matrix.Scale(1.0*mesh_scale, 4)

        #scale_matrix[0][0] = scale_matrix[0][0]*x_scalefactor
        #scale_matrix[1][1] = scale_matrix[1][1]*y_scalefactor
        #scale_matrix[2][2] = scale_matrix[2][2]*z_scalefactor
        #print("0:"+str(scale_matrix[0][0]))
        #print("1:"+str(scale_matrix[1][1]))
        #print("2:"+str(scale_matrix[2][2]))
        me_obj.dimensions.x = x_scalefactor
        me_obj.dimensions.y = y_scalefactor
        me_obj.dimensions.z = z_scalefactor
        print("0:"+str(x_scalefactor))
        print("1:"+str(y_scalefactor))
        print("2:"+str(z_scalefactor))

        #global_trans = to_blender_matrix(mesh.global_transform())
        me_obj.matrix_basis = SWAP_YZ_MATRIX * scale_matrix#* global_trans

        # apply transforms (rotation)
        bpy.ops.object.transform_apply(location=False,rotation=True,scale=True)

        # finalize mesh
        me.update(calc_edges=True)

    ##############################################
    # create bone group
    ##############################################
        if m_weights:
            for n, indices in enumerate(g_indices):
                for p,index in enumerate(indices):
                    
                    # name is index
                    bone_name = bone_list[index]
                    #print("p is:",p," index is:",index," Name is:",bone_name)
                    # use existing group or new
                    group = None
                    if bone_name in me_obj.vertex_groups.keys():
                        group = me_obj.vertex_groups[bone_name]
                    if group == None:
                        group = me_obj.vertex_groups.new(bone_name)
                    
                    # do weight                    
                    weight = m_weights[n][p]
                    if weight == 0: # no weight = no influence, so not acutally used
                        continue;
                    
                    group.add([n], weight, 'ADD')

    print("Finished Model")

    return 0 # all fine

##################################################################

def createRig(name, origin, boneTable):

    # Create armature and object
    bpy.ops.object.add(
        type='ARMATURE', 
        enter_editmode=True,
        location=origin)
    ob = bpy.context.object
    ob.show_x_ray = True
    ob.data.draw_type = 'STICK'
    ob.name = name
    amt = ob.data
    amt.name = name+'Amt'
    amt.show_axes = False
 
    # Create bones
    bpy.ops.object.mode_set(mode='EDIT')
    for (bname, pname, vector) in boneTable:    
        bone = amt.edit_bones.new(bname)
        if pname:
            parent = amt.edit_bones[pname]
            bone.parent = parent
            bone.head = parent.tail
            bone.use_connect = False
            if vector[0]+vector[1]+vector[2] < 0.0001:
                vector = (0,0.0001,0) # try to prevent zero length bones
            (trans, rot, scale) = parent.matrix.decompose()
        else:
            bone.head = (0,0,0)
            rot = Matrix.Translation((0,0,0))	# identity matrix
        bone.tail = rot * Vector(vector) + bone.head
    bpy.ops.object.mode_set(mode='OBJECT')
    return ob

def poseRig(rig, poseTable,use_layers,mesh_scale):

    SWAP_YZ_MATRIX = mathutils.Matrix.Rotation(math.radians(90.0), 4, 'X')
    bpy.context.scene.objects.active = rig
    bpy.ops.object.mode_set(mode='POSE')
 
    for (bname, loc,angle) in poseTable:
        
        pbone = rig.pose.bones[bname]

        pbone.matrix_basis.identity()
        # Set rotation mode to Euler XYZ, easier to understand
        # than default quaternions
        pbone.rotation_mode = 'XYZ'
        # Documentation bug: Euler.rotate(angle,axis):
        # axis in ['x','y','z'] and not ['X','Y','Z']
        print("quat:",angle)
        quat = Quaternion((angle[3]*-1.0,angle[0]*-1.0,angle[1],angle[2])) # flip on axis, Negate axis, negate 'w'
        euler=quat.to_euler('XYZ')
        #pbone.rotation_euler.rotate_axis('X', euler[0])
        #pbone.rotation_euler.rotate_axis('Y', euler[2])
        #pbone.rotation_euler.rotate_axis('Z', euler[1])

        #########################
        pos = Vector([float(loc[0]), float(loc[1]), float(loc[2])])
        rot = Euler([float(euler[0]), float(euler[1]), float(euler[2])])
        
        kf = Matrix.Identity(4)
        kf = Matrix.Translation(pos) * rot.to_matrix().to_4x4()

        if pbone.parent:
            pbone.matrix = pbone.parent.matrix * kf
        else:
            pbone.matrix = kf
        
    bpy.ops.pose.armature_apply()

    bpy.ops.mesh.primitive_ico_sphere_add(subdivisions=3,size=2)
    bone_vis = bpy.context.active_object
    bone_vis.data.name = bone_vis.name = "HitmanDat_bone_vis"
    bone_vis.use_fake_user = True
    bpy.context.scene.objects.unlink(bone_vis) # don't want the user deleting this
    bpy.context.scene.objects.active = rig
    ###########################

    # Calculate armature dimensions...Blender should be doing this!
    maxs = [0,0,0]
    mins = [0,0,0]
    for bone in rig.data.bones:
        for i in range(3):
            maxs[i] = max(maxs[i],bone.head_local[i])
            mins[i] = min(mins[i],bone.head_local[i])

    dimensions = []

    for i in range(3):
        dimensions.append(maxs[i] - mins[i])
        
    length = max(0.001, (dimensions[0] + dimensions[1] + dimensions[2]) / 600) # very small indeed, but a custom bone is used for display

    # Apply spheres
    bpy.ops.object.mode_set(mode='EDIT')
    for (bname, loc,angle) in poseTable:
        bone=rig.data.edit_bones[bname]
        bone.tail = bone.head + (bone.tail - bone.head).normalized() * length # Resize loose bone tails based on armature size
        rig.pose.bones[bone.name].custom_shape = bone_vis # apply bone shape

    #pbone.rotation_quaternion = quat
    bpy.ops.object.mode_set(mode='OBJECT')

    #scale
    scale_matrix = mathutils.Matrix.Scale(1.0*mesh_scale, 4)

    global_trans = mathutils.Matrix.Translation((poseTable[0][1][0]*100.0,poseTable[0][1][1]*100.0,poseTable[0][1][2]*100.0))
    matrix_basis = SWAP_YZ_MATRIX * scale_matrix * global_trans
    rig.data.transform(matrix_basis)
    
    # apply transforms (rotation)
    bpy.ops.object.transform_apply(location=True,rotation=True,scale=True)

    
    #add armature modifier
    for cnt,ob_ptr in enumerate(objects_ptr):
        mod = ob_ptr.modifiers.new('RigModifier', 'ARMATURE')
        mod.object = rig
        mod.use_bone_envelopes = False
        mod.use_vertex_groups = True

        # place armature on each layer as well
        if use_layers:
            rig.layers[cnt] = True

 
##################################################################

def alignPosition(alignment):
	alignment = alignment - 1
	filehandle.seek(((filehandle.tell() + alignment) & ~alignment), 0) # SEEK_SET

def randomColor():
    randomR = random.random()
    randomG = random.random()
    randomB = random.random()
    return (randomR, randomG, randomB)

def HalfToFloat(h):
    s = int((h >> 15) & 0x00000001)    # sign
    e = int((h >> 10) & 0x0000001f)    # exponent
    f = int(h & 0x000003ff)            # fraction

    if e == 0:
       if f == 0:
          return int(s << 31)
       else:
          while not (f & 0x00000400):
             f <<= 1
             e -= 1
          e += 1
          f &= ~0x00000400
          #print (s,e,f)
    elif e == 31:
       if f == 0:
          return int((s << 31) | 0x7f800000)
       else:
          return int((s << 31) | 0x7f800000 | (f << 13))

    e = e + (127 -15)
    f = f << 13

    return int((s << 31) | (e << 23) | f)


def file_str(long): 
   s=''
   for j in range(0,long): 
       lit =  unpack('c',filehandle.read(1))[0]
       #print("ord of "+str(lit)+" is "+str(ord(lit)))
       if ord(lit)!= 0:
           s+=lit.decode("utf-8")
       else:
           break;
   return s

def readcstr():
    buf = ''
    while True:
        b = unpack('c',filehandle.read(1))[0]
        #print("ord of "+str(b)+" is "+str(ord(b)))
        if b is None or ord(b) == 0:
            return buf
        else:
            buf+=b.decode("utf-8")

def BB(n): # Unsigned Char Default is < (little Endian)  > = Big-Endian
    array = [] 
    for id in range(n): 
        array.append(unpack('B', filehandle.read(1))[0])
    return array
def Bb(n): # Signed Char
    array = [] 
    for id in range(n): 
        array.append(unpack('b', filehandle.read(1))[0])
    return array
def BH(n): # Unsigned Short
    array = [] 
    for id in range(n): 
        array.append(unpack('H', filehandle.read(2))[0])
    return array
def Bh(n): # Signed Short
    array = [] 
    for id in range(n): 
        array.append(unpack('h', filehandle.read(2))[0])
    return array
def Bf(n): # Float
	array = [] 
	for id in range(n):
		nz=filehandle.read(4)
		#print("NZ:"+str(nz))
		array.append(unpack('f',nz )[0])
	return array
def BF(n): # HALF FLoat
    array = [] 
    for id in range(n):
        nz=filehandle.read(2)
        #print("NZ:"+str(nz))
        v = struct.unpack('H', nz)
        x = HalfToFloat(v[0])
        # hack to coerce int to float
        str = struct.pack('I',x)
        f=struct.unpack('f', str)
        array.append(f[0])
    return array
def Bi(n): # Signed Long Int
    array = [] 
    for id in range(n): 
        array.append(unpack('i', filehandle.read(4))[0])
    return array
def BI(n): # Unsigned Long Int
    array = [] 
    for id in range(n): 
        array.append(unpack('I', filehandle.read(4))[0])
    return array
def Bd(n): # Double
    array = [] 
    for id in range(n): 
        array.append(unpack('d', filehandle.read(8))[0])
    return array

def views():
    """ Returns the set of 3D views.
    """
    rtn = []
    for a in bpy.data.screens["Default"].areas:
        if a.type == 'VIEW_3D':
            rtn.append(a)
    return rtn

def import_DeusExMD(file, context, randomize_colors,import_vertcolors,skip_blank,use_layers,mesh_scale):

    global filehandle,fmt_table,model_name,header_size
    report = None
    objects_ptr = []
    header_size = 0

    time1 = time.time()
    print("start")

    workingpath=os.path.dirname(os.path.realpath(__file__))

     #adjust clipping mode
    view=views()
    print(view[0].spaces[0].type)  # should be VIEW_3D
    view[0].spaces[0].clip_end=10000

    # split off model name
    root, ext = os.path.splitext(file)    
    model_name=root.split("\\")[-1]
    print(model_name)

    ##################################

    try:
        filehandle = open(file, "rb")
    except:
        report="Error loading '"+model_name+".dat/.bin'\n"
        return report

    try:
        # first check if newer format
        flag = file_str(4)
        print("Flag:"+str(flag))
        if flag == "MIRP":
            hdr_size = BI(1)[0]
            filehandle.seek(hdr_size,1) #SEEK_CUR
            # clear some junk
            BI(4)
            header_size = hdr_size + 24
        else:
            print("fail compare")
            report=model_name+" doesn't seem to be a model file."
            return report
    except:
        filehandle.seek(0,0) #SEEK_SET
    
    #get key data about the mesh
    print("Final header size:"+str(header_size))
    end_address = BI(1)[0] + header_size

    filehandle.seek(0,2) # SEEK_END
    file_size=filehandle.tell();
    
    if file_size - end_address != 48: # end header should be at very end.
        report=model_name+" doesn't seem to be a model file."
        return report

    
    filehandle.seek(end_address,0) # SEEK_SET

    #read in junk
    BI(3)
    model_count = BI(1)[0]
    table_offset = BI(1)[0] + header_size
    print("model count:"+str(model_count))

    #read in offsets table
    block_offsets = []
    filehandle.seek(table_offset,0) # SEEK_SET
    for i in range(0,model_count):
        tmp=BI(1)
        print("Table:"+str(tmp[0]))
        block_offsets.extend([tmp[0]+ header_size])



    # unselect all   
    for item in bpy.context.selectable_objects:   
        item.select = False   

    #model_count=2 # TEMP
    # at model start
    ####################################
    for i in range(0,model_count): 
        # do mesh and materials
        result=import_mesh(context,randomize_colors,import_vertcolors,use_layers,mesh_scale,i,block_offsets[i])

        if result is None:
            report = "Mesh Output Failed."
        elif result == 1:
            report = "Vertice counts didn't match, aborting!"


    # should be at end of models, process last
    filehandle.close()
    
    #############################################
    ## Skeleton
    #############################################

    has_skeleton = True
    
    # take root filename and add .skel extension
    root = root + ".skel"
    try:
        filehandle = open(root, "rb")
    except:
        has_skeleton = False  

    if has_skeleton:
        print("HAS SKEL")
        #get key data about the mesh
        signature = BI(1)[0]
        if signature != MAGIC_SIG:
            report=model_name+".skel signature is invalid:"+str(signature)+"\n"
            return report
        
        filehandle.seek(0x18)
        bone_block_len = BI(1)[0]
        print("Bone block length:"+str(bone_block_len))

        name_start=filehandle.tell()
        filehandle.seek(bone_block_len,1) #SEEK_CUR
        
        print("ended at:"+str(filehandle.tell()))

        names_offset_len = BI(1)[0]
        
        bone_names=[] # names array

        for y in range(0,names_offset_len):
            tmp=BI(1)[0] # name offset
            offsets_start=filehandle.tell()
            filehandle.seek(name_start+tmp,0) # SEEK_SET
            bone_names.append(readcstr())
            print("name:",bone_names[y])
            filehandle.seek(offsets_start)

        
        # skip past hash list
        hashes_len = BI(1)[0]
        filehandle.seek(hashes_len*4,1) # 4 bytes per hash #SEEK_CUR

        # Now get bone parent index list
        parent_index_len = BI(1)[0]
        Bone_parents = []
        
        for y in range(0,parent_index_len):
            Bone_parents.append(Bi(1)[0])  # signed so root = -1

        #Now get location/rotation data and build skeleton table as we go
        bone_count = BI(1)[0]
        Bone_Table=[]
        Pose_Table=[]

        for y in range(0,bone_count):
            loc_n_rot = Bf(8)

            parent_name = None
            parent_id=Bone_parents[y]
            if parent_id != -1:
                parent_name = bone_names[parent_id]                

            Bone_Table.extend([(bone_names[y],parent_name,(loc_n_rot[0]*-1,loc_n_rot[1],loc_n_rot[2]))]) # Flip on X Axis
            Pose_Table.extend([(bone_names[y],(loc_n_rot[0]*-1,loc_n_rot[1],loc_n_rot[2]),(loc_n_rot[4],loc_n_rot[5],loc_n_rot[6],loc_n_rot[7]))])

        # build the skeleton
        rig = createRig('Skeleton', Vector((0,0,0)), Bone_Table)
        poseRig(rig, Pose_Table,use_layers,mesh_scale)            

        filehandle.close()

    print("time is ", time.time() - time1)
    return report
